@@ -1,5 +1,7 @@
-#!/usr/bin/env python
-import os, subprocess, sys
+#!/usr/bin/env python3
+import os
+import subprocess
+import sys
 from os.path import exists, expanduser
 from os import unlink
 
@@ -8,7 +10,7 @@ from alibuild_helpers.log import debug, banner
 
 
 def generate_analytics_id():
-  getstatusoutput("mkdir -p  ~/.config/alibuild")
+  os.makedirs(os.path.expanduser("~/.config/alibuild"), exist_ok=True)
   err, output = getstatusoutput("uuidgen >  ~/.config/alibuild/analytics-uuid")
   # If an error is found while generating the unique user ID, we disable
   # the analytics on the machine.
@@ -22,12 +24,9 @@ def askForAnalytics():
   banner("In order to improve user experience, aliBuild would like to gather "
          "analytics about your builds.\nYou can find all the details at:\n\n"
          "  https://github.com/alisw/alibuild/blob/master/ANALYTICS.md\n")
-  # raw_input and input are different between python 2 and 3
-  try: _input = raw_input
-  except NameError: _input = input
-  a = _input("Is that ok for you [YES/no]? ")
+  a = input("Is that ok for you [YES/no]? ")
   if a.strip() and a.strip().lower().startswith("n"):
-    debug("User requsted disabling analytics.")
+    debug("User requested disabling analytics.")
     return disable_analytics()
   return generate_analytics_id()
 
@@ -107,7 +106,7 @@ def report_exception(e):
     exd = e.__class__.__name__,
     exf = "1")
 
-def enable_analytics():
+def enable_analytics() -> None:
   if exists(expanduser("~/.config/alibuild/disable-analytics")):
     unlink(expanduser("~/.config/alibuild/disable-analytics"))
   if not exists(expanduser("~/.config/alibuild/analytics-uuid")):
