@@ -1,15 +1,13 @@
 import argparse
-from alibuild_helpers.utilities import format, detectArch
-from alibuild_helpers.utilities import normalise_multiple_options
+from alibuild_helpers.utilities import detectArch, normalise_multiple_options
+from alibuild_helpers.workarea import cleanup_git_log
 import multiprocessing
 
 import re
 import os
+import shlex
 
-try:
-  import commands
-except ImportError:
-  import subprocess as commands
+import subprocess as commands
 from os.path import abspath, dirname, basename
 import sys
 
@@ -24,14 +22,14 @@ DEFAULT_CHDIR = os.environ.get("ALIBUILD_CHDIR") or "."
 # --dist-tag). It can be either:
 # - A tag name
 # - A repository spec in the for org/repo@tag
-def alidist_string(s, star):
+def alidist_string(s):
   repo, have_repo_spec, ver = s.partition("@")
   if not have_repo_spec:
-    repo, ver = "alisw/%sdist" % star, s
+    repo, ver = "alisw/alidist", s
   return {"repo": repo, "ver": ver}
 
 
-def doParseArgs(star):
+def doParseArgs():
   detectedArch = detectArch()
   parser = argparse.ArgumentParser(epilog="""\
   For help about each option, specify --help after the option itself. For
@@ -67,7 +65,7 @@ def doParseArgs(star):
   build_parser.add_argument("pkgname", metavar="PACKAGE", nargs="+",
                             help="One of the packages in CONFIGDIR. May be specified multiple times.")
 
-  build_parser.add_argument("--defaults", dest="defaults", default="release", metavar="DEFAULT",
+  build_parser.add_argument("--defaults", dest="defaults", default="o2", metavar="DEFAULT",
                             help="Use defaults from CONFIGDIR/defaults-%(metavar)s.sh.")
   build_parser.add_argument("-a", "--architecture", dest="architecture", metavar="ARCH", default=detectedArch,
                             help=("Build as if on the specified architecture. When used with --docker, build "
@@ -103,6 +101,14 @@ def doParseArgs(star):
                                   "same effect as adding 'force_rebuild: true' to its recipe "
                                   "in CONFIGDIR. You can specify this option multiple times or "
                                   "separate multiple arguments with commas."))
+  build_parser.add_argument("--annotate", default=[], action="append", metavar="PACKAGE=COMMENT",
+                            help=("Store COMMENT in the build metadata for PACKAGE. This option "
+                                  "can be given multiple times, if you want to store comments "
+                                  "in multiple packages. The comment will only be stored if "
+                                  "PACKAGE is compiled or downloaded during this run; if it "
+                                  "already exists, this does not happen."))
+  build_parser.add_argument("--only-deps", dest="onlyDeps", default=False, action="store_true",
+                            help="Only build dependencies, not the main package (e.g. for caching)")
 
   build_docker = build_parser.add_argument_group(title="Build inside a container", description="""\
   Builds can be done inside a Docker container, to make it easier to get a
@@ -112,7 +118,7 @@ def doParseArgs(star):
   """)
   build_docker.add_argument("--docker", dest="docker", action="store_true",
                             help="Build inside a Docker container.")
-  build_docker.add_argument("--docker-image", dest="dockerImage", metavar="IMAGE", default=argparse.SUPPRESS,
+  build_docker.add_argument("--docker-image", dest="dockerImage", metavar="IMAGE", default=None,
                             help=("The Docker image to build inside of. Implies --docker. "
                                   "By default, an image is chosen based on the architecture."))
   build_docker.add_argument("--docker-extra-args", metavar="ARGLIST", default="",
@@ -147,14 +153,14 @@ def doParseArgs(star):
   build_remote.add_argument("--insecure", dest="insecure", action="store_true",
                             help="Don't validate TLS certificates when connecting to an https:// remote store.")
 
-  build_dirs = build_parser.add_argument_group(title="Customise %sBuild directories" % star)
+  build_dirs = build_parser.add_argument_group(title="Customise aliBuild directories")
   build_dirs.add_argument("-C", "--chdir", metavar="DIR", dest="chdir", default=DEFAULT_CHDIR,
                           help=("Change to the specified directory before building. "
                                 "Alternatively, set ALIBUILD_CHDIR. Default '%(default)s'."))
   build_dirs.add_argument("-w", "--work-dir", dest="workDir", default=DEFAULT_WORK_DIR,
                           help=("The toplevel directory under which builds should be done and build results "
                                 "should be installed. Default '%(default)s'."))
-  build_dirs.add_argument("-c", "--config-dir", dest="configDir", default="%sdist" % star,
+  build_dirs.add_argument("-c", "--config-dir", dest="configDir", default="alidist",
                           help="The directory containing build recipes. Default '%(default)s'.")
   build_dirs.add_argument("--reference-sources", dest="referenceSources", metavar="MIRRORDIR",
                           default="%(workDir)s/MIRROR",
@@ -170,8 +176,8 @@ def doParseArgs(star):
   build_system = build_parser.add_mutually_exclusive_group()
   build_system.add_argument("--always-prefer-system", dest="preferSystem", action="store_true",
                             help="Always use system packages when compatible.")
-  build_system.add_argument("--no-system", dest="noSystem", action="store_true",
-                            help="Never use system packages, even if compatible.")
+  build_system.add_argument("--no-system", dest="noSystem", nargs="?", const="*", default=None, metavar="PACKAGES",
+                            help="Never use system packages for the provided, command separated, PACKAGES, even if compatible.")
 
   # Options for clean subcommand
   clean_parser.add_argument("-a", "--architecture", dest="architecture", metavar="ARCH", default=detectedArch,
@@ -179,7 +185,7 @@ def doParseArgs(star):
                                   "architecture, which is '%(default)s'."))
   clean_parser.add_argument("--aggressive-cleanup", dest="aggressiveCleanup", action="store_true",
                             help="Delete as much build data as possible when cleaning up.")
-  clean_dirs = clean_parser.add_argument_group(title="Customise %sBuild directories" % star)
+  clean_dirs = clean_parser.add_argument_group(title="Customise aliBuild directories")
   clean_dirs.add_argument("-C", "--chdir", metavar="DIR", dest="chdir", default=DEFAULT_CHDIR,
                           help=("Change to the specified directory before cleaning up. "
                                 "Alternatively, set ALIBUILD_CHDIR. Default '%(default)s'."))
@@ -194,7 +200,7 @@ def doParseArgs(star):
                            help=("Resolve dependencies as if on the specified architecture. When used with "
                                  "--docker, use a Docker image for the specified architecture. Default is "
                                  "the current system architecture, which is '%(default)s'."))
-  deps_parser.add_argument("--defaults", dest="defaults", default="release", metavar="DEFAULT",
+  deps_parser.add_argument("--defaults", dest="defaults", default="o2", metavar="DEFAULT",
                            help="Use defaults from CONFIGDIR/defaults-%(metavar)s.sh.")
   deps_parser.add_argument("--disable", dest="disable", default=[], metavar="PACKAGE", action="append",
                            help=("Assume we're not building %(metavar)s and all its (unique) dependencies. "
@@ -216,7 +222,7 @@ def doParseArgs(star):
   """)
   deps_docker.add_argument("--docker", dest="docker", action="store_true",
                            help="Check for available system packages inside a Docker container.")
-  deps_docker.add_argument("--docker-image", dest="dockerImage", metavar="IMAGE",
+  deps_docker.add_argument("--docker-image", dest="dockerImage", metavar="IMAGE", default=None,
                            help=("The Docker image to use. Implies --docker. By default, an image "
                                  "is chosen based on the current or selected architecture."))
   deps_docker.add_argument("--docker-extra-args", default="", metavar="ARGLIST",
@@ -224,15 +230,15 @@ def doParseArgs(star):
                                  "Passed through verbatim -- separate multiple arguments "
                                  "with spaces, and make sure quoting is correct! Implies --docker."))
 
-  deps_parser.add_argument_group(title="Customise %sBuild directories" % star) \
-             .add_argument("-c", "--config-dir", dest="configDir", default="%sdist" % star,
+  deps_parser.add_argument_group(title="Customise aliBuild directories") \
+             .add_argument("-c", "--config-dir", dest="configDir", default="alidist",
                            help="The directory containing build recipes. Default '%(default)s'.")
 
   deps_system = deps_parser.add_mutually_exclusive_group()
   deps_system.add_argument("--always-prefer-system", dest="preferSystem", action="store_true",
                            help="Always use system packages when compatible.")
-  deps_system.add_argument("--no-system", dest="noSystem", action="store_true",
-                           help="Never use system packages, even if compatible.")
+  deps_system.add_argument("--no-system", dest="noSystem", nargs="?", const="*", default=None, metavar="PACKAGES",
+                           help="Never use system packages for PACKAGES, even if compatible.")
 
   # Options for the doctor subcommand
   doctor_parser.add_argument("packages", metavar="PACKAGE", nargs="+",
@@ -242,18 +248,20 @@ def doParseArgs(star):
                              help=("Resolve requirements as if on the specified architecture. When used with "
                                    "--docker, use a Docker image for the specified architecture. Default is "
                                    "the current system architecture, which is '%(default)s'."))
-  doctor_parser.add_argument("--defaults", dest="defaults", default="release", metavar="DEFAULT",
+  doctor_parser.add_argument("--defaults", dest="defaults", default="o2", metavar="DEFAULT",
                              help="Use defaults from CONFIGDIR/defaults-%(metavar)s.sh.")
   doctor_parser.add_argument("--disable", dest="disable", default=[], metavar="PACKAGE", action="append",
                              help=("Assume we're not building %(metavar)s and all its (unique) dependencies. "
                                    "You can specify this option multiple times or separate multiple arguments "
                                    "with commas."))
+  doctor_parser.add_argument("-e", dest="environment", action="append", default=[],
+                            help="KEY=VALUE binding to add to the build environment. May be specified multiple times.")
 
   doctor_system = doctor_parser.add_mutually_exclusive_group()
   doctor_system.add_argument("--always-prefer-system", dest="preferSystem", action="store_true",
                              help="Always use system packages when compatible.")
-  doctor_system.add_argument("--no-system", dest="noSystem", action="store_true",
-                             help="Never use system packages, even if compatible.")
+  doctor_system.add_argument("--no-system", dest="noSystem", nargs="?", const="*", default=None, metavar="PACKAGES",
+                             help="Never use system packages for the provided, command separated, PACKAGES, even if compatible.")
 
   doctor_docker = doctor_parser.add_argument_group(title="Use a Docker container", description="""\
   If you're planning to build inside a Docker container, e.g. using aliBuild
@@ -262,18 +270,45 @@ def doParseArgs(star):
   """)
   doctor_docker.add_argument("--docker", dest="docker", action="store_true",
                              help="Check for available system packages inside a Docker container.")
-  doctor_docker.add_argument("--docker-image", dest="dockerImage", metavar="IMAGE",
+  doctor_docker.add_argument("--docker-image", dest="dockerImage", metavar="IMAGE", default=None,
                              help=("The Docker image to use. Implies --docker. By default, an image "
                                    "is chosen based on the current or selected architecture."))
+  doctor_docker.add_argument("--docker-extra-args", metavar="ARGLIST", default="",
+                             help=("Command-line arguments to pass to 'docker run'. "
+                                   "Passed through verbatim -- separate multiple arguments "
+                                   "with spaces, and make sure quoting is correct! Implies --docker."))
 
-  doctor_dirs = doctor_parser.add_argument_group(title="Customise %sBuild directories" % star)
+  doctor_remote = doctor_parser.add_argument_group(title="Re-use prebuilt tarballs", description="""\
+  Reusing prebuilt tarballs saves compilation time, as common packages need not
+  be rebuilt from scratch. rsync://, https://, b3:// and s3:// remote stores
+  are recognised. Some of these require credentials: s3:// remotes require an
+  ~/.s3cfg; b3:// remotes require AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY
+  environment variables. A useful remote store is
+  'https://s3.cern.ch/swift/v1/alibuild-repo'. It requires no credentials and
+  provides tarballs for the most common supported architectures.
+  """)
+  doctor_remote.add_argument("--no-remote-store", action="store_true",
+                            help="Disable the use of the remote store, even if it is enabled by default.")
+  doctor_remote.add_argument("--remote-store", dest="remoteStore", metavar="STORE", default="", help="""\
+  Where to find prebuilt tarballs to reuse. See above for available remote stores.
+  End with ::rw if you want to upload (in that case, ::rw is stripped and --write-store
+  is set to the same value). Implies --no-system. May be set to a default store on some
+  architectures; use --no-remote-store to disable it in that case.
+  """)
+  doctor_remote.add_argument("--write-store", dest="writeStore", metavar="STORE", default="",
+                            help=("Where to upload newly built packages. Same syntax as --remote-store, "
+                                  "except ::rw is not recognised. Implies --no-system."))
+  doctor_remote.add_argument("--insecure", dest="insecure", action="store_true",
+                            help="Don't validate TLS certificates when connecting to an https:// remote store.")
+
+  doctor_dirs = doctor_parser.add_argument_group(title="Customise aliBuild directories")
   doctor_dirs.add_argument("-C", "--chdir", metavar="DIR", dest="chdir", default=DEFAULT_CHDIR,
                            help=("Change to the specified directory before doing anything. "
                                  "Alternatively, set ALIBUILD_CHDIR. Default '%(default)s'."))
   doctor_dirs.add_argument("-w", "--work-dir", dest="workDir", default=DEFAULT_WORK_DIR,  # TODO: previous default was "workDir".
                            help=("The toplevel directory under which builds should be done and build results "
                                  "should be installed. Default '%(default)s'."))
-  doctor_dirs.add_argument("-c", "--config", dest="configDir", default="%sdist" % star,
+  doctor_dirs.add_argument("-c", "--config", dest="configDir", default="alidist",
                            help="The directory containing build recipes. Default '%(default)s'.")
 
   # Options for the init subcommand
@@ -283,27 +318,27 @@ def doParseArgs(star):
                            help=("Parse defaults using the specified architecture. Default is "
                                  "the current system architecture, which is '%(default)s'."))
 
-  init_parser.add_argument("--defaults", dest="defaults", default="release", metavar="DEFAULT",
+  init_parser.add_argument("--defaults", dest="defaults", default="o2", metavar="DEFAULT",
                            help="Use defaults from CONFIGDIR/defaults-%(metavar)s.sh.")
   init_parser.add_argument("-z", "--devel-prefix", dest="develPrefix", default=".",
                            help=("Directory under which to clone the repository of build recipes. "
                                  "See also: -c/--config-dir. Default '%(default)s'."))
 
   init_parser.add_argument("--dist", metavar="[USER/REPO@]BRANCH", dest="dist", default="",
-                           type=lambda x: alidist_string(x, star),
+                           type=alidist_string,
                            help=("Download the given repository containing build recipes into "
                                  "CONFIGDIR. Syntax: [user/repo@]branch or [url@]branch. The "
-                                 "default repo is 'alisw/%sdist; the default branch is the "
-                                 "repository's main branch." % star))
+                                 "default repo is 'alisw/alidist; the default branch is the "
+                                 "repository's main branch."))
 
-  init_dirs = init_parser.add_argument_group(title="Customise %sBuild directories" % star)
+  init_dirs = init_parser.add_argument_group(title="Customise aliBuild directories")
   init_dirs.add_argument("-C", "--chdir", metavar="DIR", dest="chdir", default=DEFAULT_CHDIR,
                          help=("Change to the specified directory before doing anything. "
                                "Alternatively, set ALIBUILD_CHDIR. Default '%(default)s'."))
   init_dirs.add_argument("-w", "--work-dir", dest="workDir", default=DEFAULT_WORK_DIR,
                          help=("The toplevel directory under which builds should be done and "
                                "build results should be installed. Default '%(default)s'."))
-  init_dirs.add_argument("-c", "--config-dir", dest="configDir", default="%%(prefix)s%sdist" % star,
+  init_dirs.add_argument("-c", "--config-dir", dest="configDir", default="%(prefix)salidist",
                          help=("The directory where build recipes will be placed. '%%(prefix)s' will "
                                "be replaced with 'DEVELPREFIX/'. Default '%(default)s'."))
   init_dirs.add_argument("--reference-sources", dest="referenceSources", metavar="MIRRORDIR",
@@ -327,10 +362,10 @@ def doParseArgs(star):
     return 2
   rest.sort(key=optionOrder)
   sys.argv = [prog] + rest
-  args = finaliseArgs(parser.parse_args(), parser, star)
+  args = finaliseArgs(parser.parse_args(), parser)
   return (args, parser)
 
-VALID_ARCHS_RE = "^slc[5-9]_(x86-64|ppc64)$|^(ubuntu|ubt|osx|fedora)[0-9]*_(x86-64|arm64)$"
+VALID_ARCHS_RE = "^slc[5-9]_(x86-64|ppc64|aarch64)$|^(ubuntu|ubt|osx|fedora)[0-9]*_(x86-64|arm64)$"
 
 def matchValidArch(architecture):
   return bool(re.match(VALID_ARCHS_RE, architecture))
@@ -340,22 +375,28 @@ On Linux, x86-64:
    RHEL6 / SLC6 compatible: slc6_x86-64
    RHEL7 / CC7 compatible: slc7_x86-64
    RHEL8 / CC8 compatible: slc8_x86-64
-   Ubuntu 18.04 compatible: ubuntu1804_x86-64
+   RHEL9 / ALMA9 compatible: slc9_x86-64
    Ubuntu 20.04 compatible: ubuntu2004_x86-64
+   Ubuntu 22.04 compatible: ubuntu2204_x86-64
+   Ubuntu 24.04 compatible: ubuntu2404_x86-64
    Fedora 33 compatible: fedora33_x86-64
    Fedora 34 compatible: fedora34_x86-64
 
+On Linux, ARM:
+   RHEL9 / ALMA9 compatible: slc9_aarch64
+
 On Linux, POWER8 / PPC64 (little endian):
    RHEL7 / CC7 compatible: slc7_ppc64
 
-On Mac, x86-64:
-   Yosemite to Big Sur: osx_x86-64
-   Big Sur: osx_arm64
+On Mac, 1-2 latest supported OSX versions:
+   Intel: osx_x86-64
+   Apple Silicon: osx_arm64
 """
 
-S3_SUPPORTED_ARCHS = "slc7_x86-64", "slc8_x86-64", "ubuntu2004_x86-64"
+# When updating this variable, also update docs/docs/user.md!
+S3_SUPPORTED_ARCHS = "slc7_x86-64", "slc8_x86-64", "ubuntu2004_x86-64", "ubuntu2204_x86-64", "ubuntu2404_x86-64", "slc9_x86-64", "slc9_aarch64"
 
-def finaliseArgs(args, parser, star):
+def finaliseArgs(args, parser):
 
   # Nothing to finalise for version or analytics
   if args.action in ["version", "analytics", "architecture"]:
@@ -379,25 +420,18 @@ def finaliseArgs(args, parser, star):
     args.force_rebuild = normalise_multiple_options(args.force_rebuild)
 
   if args.action in ["build", "init"]:
-    args.referenceSources = format(args.referenceSources, workDir=args.workDir)
+    args.referenceSources = args.referenceSources % {"workDir": args.workDir}
+    # Do this cleanup as early as possible to avoid false positives due to
+    # stale git logs from previous invocations.
+    cleanup_git_log(args.referenceSources)
 
-  if args.action == "build":
-    args.configDir = args.configDir
-
-    # On selected platforms, caching is active by default
-    if args.architecture in S3_SUPPORTED_ARCHS and not args.preferSystem and not args.no_remote_store:
-      args.noSystem = True
-      if not args.remoteStore:
-        args.remoteStore = "https://s3.cern.ch/swift/v1/alibuild-repo"
-    elif args.no_remote_store:
-      args.remoteStore = ""
-
-    if args.remoteStore or args.writeStore:
-      args.noSystem = True
-
-    if "dockerImage" in args or args.docker_extra_args:
+  if args.action in ("build", "doctor", "deps"):
+    if args.dockerImage or args.docker_extra_args:
       args.docker = True
 
+    args.docker_extra_args = shlex.split(args.docker_extra_args)
+    args.docker_extra_args.append("--network=host")
+
     if args.docker and args.architecture.startswith("osx"):
       parser.error("cannot use `-a %s` and --docker" % args.architecture)
 
@@ -407,8 +441,32 @@ def finaliseArgs(args, parser, star):
     # If specified, used the docker image requested, otherwise, if running
     # in docker the docker image is given by the first part of the
     # architecture we want to build for.
-    if args.docker and not "dockerImage" in args:
-      args.dockerImage = "alisw/%s-builder" % args.architecture.split("_")[0]
+    if args.docker and not args.dockerImage:
+      args.dockerImage = "registry.cern.ch/alisw/%s-builder" % args.architecture.split("_")[0]
+
+  if "annotate" in args:
+    for comment_assignment in args.annotate:
+      if "=" not in comment_assignment:
+        parser.error("--annotate takes arguments of the form PACKAGE=COMMENT")
+    args.annotate = {
+      package: comment
+      for package, _, comment
+      in (assignment.partition("=") for assignment in args.annotate)
+    }
+
+  if args.action in ("build", "doctor"):
+    args.configDir = args.configDir
+
+    # On selected platforms, caching is active by default
+    if args.architecture in S3_SUPPORTED_ARCHS and not args.preferSystem and not args.no_remote_store:
+      args.noSystem = "*"
+      if not args.remoteStore:
+        args.remoteStore = "https://s3.cern.ch/swift/v1/alibuild-repo"
+    elif args.no_remote_store:
+      args.remoteStore = ""
+
+    if args.remoteStore or args.writeStore:
+      args.noSystem = "*"
 
     if args.remoteStore.endswith("::rw") and args.writeStore:
       parser.error("cannot specify ::rw and --write-store at the same time")
@@ -418,16 +476,16 @@ def finaliseArgs(args, parser, star):
       args.writeStore = args.remoteStore
 
   if args.action in ["build", "init"]:
-    if "develPrefix" in args and args.develPrefix == None:
+    if "develPrefix" in args and args.develPrefix is None:
       if "chdir" in args:
         args.develPrefix = basename(abspath(args.chdir))
       else:
         args.develPrefix = basename(dirname(abspath(args.configDir)))
-    if "dockerImage" in args:
+    if getattr(args, "docker", False):
       args.develPrefix = "%s-%s" % (args.develPrefix, args.architecture) if "develPrefix" in args else args.architecture
 
   if args.action == "init":
-    args.configDir = format(args.configDir, prefix=args.develPrefix+"/")
+    args.configDir = args.configDir % {"prefix": args.develPrefix + "/"}
   elif args.action == "build":
     pass
   elif args.action == "clean":
