@@ -8,23 +8,39 @@ import sys
 import time
 import requests
 from requests.exceptions import RequestException
+from urllib.parse import quote
 
 from alibuild_helpers.cmd import execute
-from alibuild_helpers.log import debug, warning, error, dieOnError
-from alibuild_helpers.utilities import format, star, resolve_store_path, resolve_links_path
+from alibuild_helpers.log import debug, info, error, dieOnError, ProgressPrint
+from alibuild_helpers.utilities import resolve_store_path, resolve_links_path, symlink
+
+
+def remote_from_url(read_url, write_url, architecture, work_dir, insecure=False):
+  """Parse remote store URLs and return the correct RemoteSync instance for them."""
+  if read_url.startswith("http"):
+    return HttpRemoteSync(read_url, architecture, work_dir, insecure)
+  if read_url.startswith("s3://"):
+    return S3RemoteSync(read_url, write_url, architecture, work_dir)
+  if read_url.startswith("b3://"):
+    return Boto3RemoteSync(read_url, write_url, architecture, work_dir)
+  if read_url.startswith("cvmfs://"):
+    return CVMFSRemoteSync(read_url, None, architecture, work_dir)
+  if read_url:
+    return RsyncRemoteSync(read_url, write_url, architecture, work_dir)
+  return NoRemoteSync()
 
 
 class NoRemoteSync:
   """Helper class which does not do anything to sync"""
-  def syncToLocal(self, p, spec):
+  def fetch_symlinks(self, spec) -> None:
     pass
-  def syncToRemote(self, p, spec):
+  def fetch_tarball(self, spec) -> None:
     pass
-  def syncDistLinksToRemote(self, link_dir):
+  def upload_symlinks_and_tarball(self, spec) -> None:
     pass
 
 class PartialDownloadError(Exception):
-  def __init__(self, downloaded, size):
+  def __init__(self, downloaded, size) -> None:
     self.downloaded = downloaded
     self.size = size
   def __str__(self):
@@ -32,7 +48,7 @@ class PartialDownloadError(Exception):
 
 
 class HttpRemoteSync:
-  def __init__(self, remoteStore, architecture, workdir, insecure):
+  def __init__(self, remoteStore, architecture, workdir, insecure) -> None:
     self.remoteStore = remoteStore
     self.writeStore = ""
     self.architecture = architecture
@@ -42,8 +58,9 @@ class HttpRemoteSync:
     self.httpConnRetries = 4
     self.httpBackoff = 0.4
 
-  def getRetry(self, url, dest=None, returnResult=False, log=True, session=None):
+  def getRetry(self, url, dest=None, returnResult=False, log=True, session=None, progress=debug):
     get = session.get if session is not None else requests.get
+    url = quote(url, safe=":/")
     for i in range(0, self.httpConnRetries):
       if i > 0:
         pauseSec = self.httpBackoff * (2 ** (i - 1))
@@ -78,9 +95,9 @@ class HttpRemoteSync:
               if log and size != -1:
                 now = time.time()
                 if downloaded == size:
-                  debug("Download complete")
-                elif now - reportTime > 3:
-                  debug("%.0f%% downloaded...", 100*downloaded/size)
+                  progress("[100%%] Download complete")
+                elif now - reportTime > 1:
+                  progress("[%.0f%%] downloaded...", 100 * downloaded / size)
                   reportTime = now
           finally:
             if destFp:
@@ -122,7 +139,7 @@ class HttpRemoteSync:
             pass
     return None
 
-  def syncToLocal(self, p, spec):
+  def fetch_tarball(self, spec) -> None:
     # Check for any existing tarballs we can use instead of fetching new ones.
     for pkg_hash in spec["remote_hashes"]:
       try:
@@ -137,12 +154,12 @@ class HttpRemoteSync:
             arch=re.escape(self.architecture),
         ), os.path.basename(tarball)):
           debug("Previously downloaded tarball for %s with hash %s, reusing",
-                p, pkg_hash)
+                spec["package"], pkg_hash)
           return
 
     with requests.Session() as session:
       debug("Updating remote store for package %s; trying hashes %s",
-            p, ", ".join(spec["remote_hashes"]))
+            spec["package"], ", ".join(spec["remote_hashes"]))
       store_path = use_tarball = None
       # Find the first tarball that matches any possible hash and fetch it.
       for pkg_hash in spec["remote_hashes"]:
@@ -154,19 +171,35 @@ class HttpRemoteSync:
           break
 
       if store_path is None or use_tarball is None:
-        debug("Nothing fetched for %s (%s)", p, ", ".join(spec["remote_hashes"]))
+        debug("Nothing fetched for %s (%s)", spec["package"],
+              ", ".join(spec["remote_hashes"]))
         return
 
-      links_path = resolve_links_path(self.architecture, spec["package"])
-      execute("mkdir -p {}/{} {}/{}".format(self.workdir, store_path,
-                                            self.workdir, links_path))
+      os.makedirs(os.path.join(self.workdir, store_path), exist_ok=True)
 
       destPath = os.path.join(self.workdir, store_path, use_tarball)
-      if not os.path.isfile(destPath):
-        # Do not download twice
+      if not os.path.isfile(destPath):   # do not download twice
+        progress = ProgressPrint("Downloading tarball for %s@%s" %
+                                 (spec["package"], spec["version"]), min_interval=5.0)
+        progress("[0%%] Starting download of %s", use_tarball)  # initialise progress bar
         self.getRetry("/".join((self.remoteStore, store_path, use_tarball)),
-                      destPath, session=session)
+                      destPath, session=session, progress=progress)
+        progress.end("done")
 
+  def fetch_symlinks(self, spec) -> None:
+    links_path = resolve_links_path(self.architecture, spec["package"])
+    os.makedirs(os.path.join(self.workdir, links_path), exist_ok=True)
+
+    # If we already have a symlink we can use, don't update the list. This
+    # speeds up rebuilds significantly.
+    if any(f"/{pkg_hash[:2]}/{pkg_hash}/" in target
+           for target in (os.readlink(os.path.join(self.workdir, links_path, link))
+                          for link in os.listdir(os.path.join(self.workdir, links_path)))
+           for pkg_hash in spec["remote_hashes"]):
+      debug("Found symlink for %s@%s, not updating", spec["package"], spec["version"])
+      return
+
+    with requests.Session() as session:
       # Fetch manifest file with initial symlinks. This file is updated
       # regularly; we use it to avoid many small network requests.
       manifest = self.getRetry("%s/%s.manifest" % (self.remoteStore, links_path),
@@ -177,10 +210,6 @@ class HttpRemoteSync:
                                       for line in manifest.splitlines())
         if sep and linkname and target
       }
-      # If we've just downloaded a tarball, add a symlink to it.
-      # We need to strip the leading TARS/ first, though.
-      assert store_path.startswith("TARS/"), store_path
-      symlinks[use_tarball] = os.path.join(store_path[len("TARS/"):], use_tarball)
       # Now add any remaining symlinks that aren't in the manifest yet. There
       # should always be relatively few of these, as the separate network
       # requests are a bit expensive.
@@ -202,35 +231,30 @@ class HttpRemoteSync:
                           returnResult=True, log=False, session=session) \
                 .decode("utf-8").rstrip("\r\n")
     for linkname, target in symlinks.items():
-      execute("ln -nsf ../../{target} {workdir}/{linkdir}/{name}".format(
-        workdir=self.workdir, linkdir=links_path, name=linkname,
-        target=target))
+      symlink("../../" + target.lstrip("./"),
+              os.path.join(self.workdir, links_path, linkname))
 
-  def syncToRemote(self, p, spec):
-    pass
-
-  def syncDistLinksToRemote(self, link_dir):
+  def upload_symlinks_and_tarball(self, spec) -> None:
     pass
 
 
 class RsyncRemoteSync:
   """Helper class to sync package build directory using RSync."""
 
-  def __init__(self, remoteStore, writeStore, architecture, workdir):
+  def __init__(self, remoteStore, writeStore, architecture, workdir) -> None:
     self.remoteStore = re.sub("^ssh://", "", remoteStore)
     self.writeStore = re.sub("^ssh://", "", writeStore)
     self.architecture = architecture
     self.workdir = workdir
 
-  def syncToLocal(self, p, spec):
-    debug("Updating remote store for package %s with hashes %s", p,
+  def fetch_tarball(self, spec) -> None:
+    info("Downloading tarball for %s@%s, if available", spec["package"], spec["version"])
+    debug("Updating remote store for package %s with hashes %s", spec["package"],
           ", ".join(spec["remote_hashes"]))
     err = execute("""\
-    mkdir -p {workDir}/{linksPath}
-    rsync -rlvW --delete {remoteStore}/{linksPath}/ {workDir}/{linksPath}/ || :
     for storePath in {storePaths}; do
       # Only get the first matching tarball. If there are multiple with the
-      # same hash, we only need one and they should be interchangable.
+      # same hash, we only need one and they should be interchangeable.
       if tars=$(rsync -s --list-only "{remoteStore}/$storePath/{pkg}-{ver}-*.{arch}.tar.gz" 2>/dev/null) &&
          # Strip away the metadata in rsync's file listing, leaving only the first filename.
          tar=$(echo "$tars" | sed -rn '1s#[- a-z0-9,/]* [0-9]{{2}}:[0-9]{{2}}:[0-9]{{2}} ##p') &&
@@ -244,37 +268,114 @@ class RsyncRemoteSync:
         break
       fi
     done
-    """.format(pkg=p, ver=spec["version"], arch=self.architecture,
+    """.format(pkg=spec["package"], ver=spec["version"], arch=self.architecture,
                remoteStore=self.remoteStore,
                workDir=self.workdir,
-               linksPath=resolve_links_path(self.architecture, p),
                storePaths=" ".join(resolve_store_path(self.architecture, pkg_hash)
                                    for pkg_hash in spec["remote_hashes"])))
-    dieOnError(err, "Unable to update from specified store.")
+    dieOnError(err, "Unable to fetch tarball from specified store.")
 
-  def syncToRemote(self, p, spec):
+  def fetch_symlinks(self, spec) -> None:
+    links_path = resolve_links_path(self.architecture, spec["package"])
+    os.makedirs(os.path.join(self.workdir, links_path), exist_ok=True)
+    err = execute("rsync -rlvW --delete {remote_store}/{links_path}/ {workdir}/{links_path}/".format(
+      remote_store=self.remoteStore,
+      links_path=links_path,
+      workdir=self.workdir,
+    ))
+    dieOnError(err, "Unable to fetch symlinks from specified store.")
+
+  def upload_symlinks_and_tarball(self, spec) -> None:
     if not self.writeStore:
       return
-    tarballNameWithRev = format("%(package)s-%(version)s-%(revision)s.%(architecture)s.tar.gz",
-                                architecture=self.architecture,
-                                **spec)
-    cmd = format("cd %(workdir)s && "
-                 "rsync -avR --ignore-existing %(storePath)s/%(tarballNameWithRev)s  %(remoteStore)s/ &&"
-                 "rsync -avR --ignore-existing %(linksPath)s/%(tarballNameWithRev)s  %(remoteStore)s/",
-                 workdir=self.workdir,
-                 remoteStore=self.remoteStore,
-                 storePath=resolve_store_path(self.architecture, spec["hash"]),
-                 linksPath=resolve_links_path(self.architecture, p),
-                 tarballNameWithRev=tarballNameWithRev)
-    err = execute(cmd)
-    dieOnError(err, "Unable to upload tarball.")
+    dieOnError(execute("""\
+    set -e
+    cd {workdir}
+    tarball={package}-{version}-{revision}.{arch}.tar.gz
+    rsync -avR --ignore-existing "{links_path}/$tarball" {remote}/
+    for link_dir in dist dist-direct dist-runtime; do
+      rsync -avR --ignore-existing "TARS/{arch}/$link_dir/{package}/{package}-{version}-{revision}/" {remote}/
+    done
+    rsync -avR --ignore-existing "{store_path}/$tarball" {remote}/
+    """.format(
+      workdir=self.workdir,
+      remote=self.remoteStore,
+      store_path=resolve_store_path(self.architecture, spec["hash"]),
+      links_path=resolve_links_path(self.architecture, spec["package"]),
+      arch=self.architecture,
+      package=spec["package"],
+      version=spec["version"],
+      revision=spec["revision"],
+    )), "Unable to upload tarball.")
 
-  def syncDistLinksToRemote(self, link_dir):
-    if not self.writeStore:
-      return
-    execute("cd {w} && rsync -avR --ignore-existing {t}/ {rs}/".format(
-      w=self.workdir, rs=self.writeStore, t=link_dir))
+class CVMFSRemoteSync:
+  """ Sync packages build directory from CVMFS or similar
+      FS based deployment. The tarball will be created on the fly with a single
+      symlink to the remote store in it, so that unpacking really
+      means unpacking the symlink to the wanted package.
+  """
 
+  def __init__(self, remoteStore, writeStore, architecture, workdir) -> None:
+    self.remoteStore = re.sub("^cvmfs://", "", remoteStore)
+    # We do not support uploading directly to CVMFS, for obvious
+    # reasons.
+    assert(writeStore is None)
+    self.writeStore = None
+    self.architecture = architecture
+    self.workdir = workdir
+
+  def fetch_tarball(self, spec) -> None:
+    info("Downloading tarball for %s@%s-%s, if available", spec["package"], spec["version"], spec["revision"])
+    # If we already have a tarball with any equivalent hash, don't check S3.
+    for pkg_hash in spec["remote_hashes"] + spec["local_hashes"]:
+      store_path = resolve_store_path(self.architecture, pkg_hash)
+      pattern = os.path.join(self.workdir, store_path, "%s-*.tar.gz" % spec["package"])
+      if glob.glob(pattern):
+        info("Reusing existing tarball for %s@%s", spec["package"], pkg_hash)
+        return
+    info("Could not find prebuilt tarball for %s@%s-%s, will be rebuilt",
+         spec["package"], spec["version"], spec["revision"])
+
+  def fetch_symlinks(self, spec) -> None:
+    # When using CVMFS, we create the symlinks grass by reading the .
+    info("Fetching available build hashes for %s, from %s", spec["package"], self.remoteStore)
+    links_path = resolve_links_path(self.architecture, spec["package"])
+    os.makedirs(os.path.join(self.workdir, links_path), exist_ok=True)
+
+    cvmfs_architecture = re.sub(r"slc(\d+)_x86-64", r"el\1-x86_64", self.architecture)
+    err = execute("""\
+    set -x
+    # Exit without error in case we do not have any package published
+    test -d "{remote_store}/{cvmfs_architecture}/Packages/{package}" || exit 0
+    mkdir -p "{workDir}/{links_path}"
+    for install_path in $(find "{remote_store}/{cvmfs_architecture}/Packages/{package}" -type d -mindepth 1 -maxdepth 1); do
+      full_version="${{install_path##*/}}"
+      tarball={package}-$full_version.{architecture}.tar.gz
+      pkg_hash=$(cat "${{install_path}}/.build-hash" || jq -r '.package.hash' <${{install_path}}/.meta.json)
+      if [ "X$pkg_hash" = X ]; then
+        continue
+      fi
+      ln -sf ../../{architecture}/store/${{pkg_hash:0:2}}/$pkg_hash/$tarball "{workDir}/{links_path}/$tarball"
+      # Create the dummy tarball, if it does not exists
+      test -f "{workDir}/{architecture}/store/${{pkg_hash:0:2}}/$pkg_hash/$tarball" && continue
+      mkdir -p "{workDir}/INSTALLROOT/$pkg_hash/{architecture}/{package}"
+      find "{remote_store}/{cvmfs_architecture}/Packages/{package}/$full_version" ! -name etc -maxdepth 1 -mindepth 1 -exec ln -sf {} "{workDir}/INSTALLROOT/$pkg_hash/{architecture}/{package}/" \\;
+      cp -fr "{remote_store}/{cvmfs_architecture}/Packages/{package}/$full_version/etc" "{workDir}/INSTALLROOT/$pkg_hash/{architecture}/{package}/etc"
+      mkdir -p "{workDir}/TARS/{architecture}/store/${{pkg_hash:0:2}}/$pkg_hash"
+      tar -C "{workDir}/INSTALLROOT/$pkg_hash" -czf "{workDir}/TARS/{architecture}/store/${{pkg_hash:0:2}}/$pkg_hash/$tarball" .
+      rm -rf "{workDir}/INSTALLROOT/$pkg_hash"
+    done
+    """.format(
+      workDir=self.workdir,
+      architecture=self.architecture,
+      cvmfs_architecture=cvmfs_architecture,
+      package=spec["package"],
+      remote_store=self.remoteStore,
+      links_path=links_path,
+    ))
+
+  def upload_symlinks_and_tarball(self, spec) -> None:
+    dieOnError(True, "CVMFS backend does not support uploading directly")
 
 class S3RemoteSync:
   """Sync package build directory from and to S3 using s3cmd.
@@ -282,15 +383,16 @@ class S3RemoteSync:
   s3cmd must be installed separately in order for this to work.
   """
 
-  def __init__(self, remoteStore, writeStore, architecture, workdir):
+  def __init__(self, remoteStore, writeStore, architecture, workdir) -> None:
     self.remoteStore = re.sub("^s3://", "", remoteStore)
     self.writeStore = re.sub("^s3://", "", writeStore)
     self.architecture = architecture
     self.workdir = workdir
 
-  def syncToLocal(self, p, spec):
-    debug("Updating remote store for package %s with hashes %s", p,
-          ", ".join(spec["remote_hashes"]))
+  def fetch_tarball(self, spec) -> None:
+    info("Downloading tarball for %s@%s, if available", spec["package"], spec["version"])
+    debug("Updating remote store for package %s with hashes %s",
+          spec["package"], ", ".join(spec["remote_hashes"]))
     err = execute("""\
     for storePath in {storePaths}; do
       # For the first store path that contains tarballs, fetch them, and skip
@@ -302,6 +404,16 @@ class S3RemoteSync:
         break
       fi
     done
+    """.format(
+      workDir=self.workdir,
+      b=self.remoteStore,
+      storePaths=" ".join(resolve_store_path(self.architecture, pkg_hash)
+                          for pkg_hash in spec["remote_hashes"]),
+    ))
+    dieOnError(err, "Unable to fetch tarball from specified store.")
+
+  def fetch_symlinks(self, spec) -> None:
+    err = execute("""\
     mkdir -p "{workDir}/{linksPath}"
     find "{workDir}/{linksPath}" -type l -delete
     curl -sL "https://s3.cern.ch/swift/v1/{b}/{linksPath}.manifest" |
@@ -317,47 +429,51 @@ class S3RemoteSync:
     done
     """.format(
       b=self.remoteStore,
-      storePaths=" ".join(resolve_store_path(self.architecture, pkg_hash)
-                          for pkg_hash in spec["remote_hashes"]),
-      linksPath=resolve_links_path(self.architecture, p),
+      linksPath=resolve_links_path(self.architecture, spec["package"]),
       workDir=self.workdir,
     ))
-    dieOnError(err, "Unable to update from specified store.")
+    dieOnError(err, "Unable to fetch symlinks from specified store.")
 
-  def syncToRemote(self, p, spec):
+  def upload_symlinks_and_tarball(self, spec) -> None:
     if not self.writeStore:
       return
-    tarballNameWithRev = format("%(package)s-%(version)s-%(revision)s.%(architecture)s.tar.gz",
-                                architecture=self.architecture,
-                                **spec)
-    cmd = format("cd %(workdir)s && "
-                 "TARSHA256=`sha256sum %(storePath)s/%(tarballNameWithRev)s | awk '{ print $1 }'` && "
-                 "s3cmd put -s -v --host s3.cern.ch --host-bucket %(b)s.s3.cern.ch %(storePath)s/%(tarballNameWithRev)s s3://%(b)s/%(storePath)s/ 2>&1 || true\n"
-                 "HASHEDURL=`readlink %(linksPath)s/%(tarballNameWithRev)s | sed -e's|^../../||'` && "
-                 "echo $HASHEDURL | s3cmd put -s -v --host s3.cern.ch --host-bucket %(b)s.s3.cern.ch - s3://%(b)s/%(linksPath)s/%(tarballNameWithRev)s 2>&1 || true\n",
-                 workdir=self.workdir,
-                 b=self.remoteStore,
-                 storePath=resolve_store_path(self.architecture, spec["hash"]),
-                 linksPath=resolve_links_path(self.architecture, p),
-                 tarballNameWithRev=tarballNameWithRev)
-    err = execute(cmd)
-    dieOnError(err, "Unable to upload tarball.")
+    dieOnError(execute("""\
+    set -e
+    put () {{
+      s3cmd put -s -v --host s3.cern.ch --host-bucket {bucket}.s3.cern.ch "$@" 2>&1
+    }}
+    tarball={package}-{version}-{revision}.{arch}.tar.gz
+    cd {workdir}
 
-  def syncDistLinksToRemote(self, link_dir):
-    if not self.writeStore:
-      return
-    execute("""\
-    cd {w} || exit 1
-    find {t} -type l | while read -r x; do
-      hashedurl=$(readlink "$x" | sed 's|.*/[.][.]/TARS|TARS|') || exit 1
-      echo $hashedurl |
-        s3cmd put --skip-existing -q -P -s \
-                  --add-header="x-amz-website-redirect-location:\
-https://s3.cern.ch/swift/v1/{b}/$hashedurl" \
-                  --host s3.cern.ch --host-bucket {b}.s3.cern.ch \
-                  - "s3://{b}/$x" 2>&1
+    # First, upload "main" symlink, to reserve this revision number, in case
+    # the below steps fail.
+    readlink "{links_path}/$tarball" | sed 's|^\\.\\./\\.\\./||' |
+      put - "s3://{bucket}/{links_path}/$tarball"
+
+    # Then, upload dist symlink trees -- these must be in place before the main
+    # tarball.
+    find TARS/{arch}/{{dist,dist-direct,dist-runtime}}/{package}/{package}-{version}-{revision}/ \
+         -type l | while read -r link; do
+      hashedurl=$(readlink "$link" | sed 's|.*/\\.\\./TARS|TARS|')
+      echo "$hashedurl" |
+        put --skip-existing -q -P \\
+            --add-header="x-amz-website-redirect-location:\
+https://s3.cern.ch/swift/v1/{bucket}/$hashedurl" \\
+            - "s3://{bucket}/$link" 2>&1
     done
-    """.format(w=self.workdir, b=self.writeStore, t=link_dir))
+
+    # Finally, upload the tarball.
+    put "{store_path}/$tarball" s3://{bucket}/{store_path}/
+    """.format(
+      workdir=self.workdir,
+      bucket=self.remoteStore,
+      store_path=resolve_store_path(self.architecture, spec["hash"]),
+      links_path=resolve_links_path(self.architecture, spec["package"]),
+      arch=self.architecture,
+      package=spec["package"],
+      version=spec["version"],
+      revision=spec["revision"],
+    )), "Unable to upload tarball.")
 
 
 class Boto3RemoteSync:
@@ -372,14 +488,14 @@ class Boto3RemoteSync:
   time.
   """
 
-  def __init__(self, remoteStore, writeStore, architecture, workdir):
+  def __init__(self, remoteStore, writeStore, architecture, workdir) -> None:
     self.remoteStore = re.sub("^b3://", "", remoteStore)
     self.writeStore = re.sub("^b3://", "", writeStore)
     self.architecture = architecture
     self.workdir = workdir
     self._s3_init()
 
-  def _s3_init(self):
+  def _s3_init(self) -> None:
     # This is a separate method so that we can patch it out for unit tests.
     # Import boto3 here, so that if we don't use this remote store, we don't
     # have to install it in the first place.
@@ -395,7 +511,7 @@ class Boto3RemoteSync:
                              aws_secret_access_key=os.environ["AWS_SECRET_ACCESS_KEY"])
     except KeyError:
       error("you must pass the AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY env "
-            "variables to %sBuild in order to use the S3 remote store", star())
+            "variables to aliBuild in order to use the S3 remote store")
       sys.exit(1)
 
   def _s3_listdir(self, dirname):
@@ -416,23 +532,18 @@ class Boto3RemoteSync:
       raise
     return True
 
-  def syncToLocal(self, p, spec):
-    from botocore.exceptions import ClientError
-    debug("Updating remote store for package %s with hashes %s", p,
+  def fetch_tarball(self, spec) -> None:
+    debug("Updating remote store for package %s with hashes %s", spec["package"],
           ", ".join(spec["remote_hashes"]))
 
     # If we already have a tarball with any equivalent hash, don't check S3.
-    have_tarball = False
     for pkg_hash in spec["remote_hashes"]:
       store_path = resolve_store_path(self.architecture, pkg_hash)
-      if glob.glob(os.path.join(self.workdir, store_path, "%s-*.tar.gz" % p)):
-        debug("Reusing existing tarball for %s@%s", p, pkg_hash)
-        have_tarball = True
-        break
+      if glob.glob(os.path.join(self.workdir, store_path, "%s-*.tar.gz" % spec["package"])):
+        debug("Reusing existing tarball for %s@%s", spec["package"], pkg_hash)
+        return
 
     for pkg_hash in spec["remote_hashes"]:
-      if have_tarball:
-        break
       store_path = resolve_store_path(self.architecture, pkg_hash)
 
       # We don't already have a tarball with the hash that we need, so download
@@ -441,19 +552,27 @@ class Boto3RemoteSync:
       # ever use one anyway.)
       for tarball in self._s3_listdir(store_path):
         debug("Fetching tarball %s", tarball)
+        progress = ProgressPrint("Downloading tarball for %s@%s" %
+                                 (spec["package"], spec["version"]), min_interval=5.0)
+        progress("[0%%] Starting download of %s", tarball)   # initialise progress bar
         # Create containing directory locally. (exist_ok= is python3-specific.)
         os.makedirs(os.path.join(self.workdir, store_path), exist_ok=True)
-        self.s3.download_file(Bucket=self.remoteStore, Key=tarball,
-                              Filename=os.path.join(self.workdir, store_path,
-                                                    os.path.basename(tarball)))
-        have_tarball = True  # break out of outer loop
-        break
+        meta = self.s3.head_object(Bucket=self.remoteStore, Key=tarball)
+        total_size = int(meta.get("ContentLength", 0))
+        self.s3.download_file(
+          Bucket=self.remoteStore, Key=tarball,
+          Filename=os.path.join(self.workdir, store_path, os.path.basename(tarball)),
+          Callback=lambda num_bytes: progress("[%d/%d] bytes transferred", num_bytes, total_size),
+        )
+        progress.end("done")
+        return
 
-    if not have_tarball:
-      debug("Remote has no tarballs for %s with hashes %s", p,
-            ", ".join(spec["remote_hashes"]))
+    debug("Remote has no tarballs for %s with hashes %s", spec["package"],
+          ", ".join(spec["remote_hashes"]))
 
-    links_path = resolve_links_path(self.architecture, p)
+  def fetch_symlinks(self, spec) -> None:
+    from botocore.exceptions import ClientError
+    links_path = resolve_links_path(self.architecture, spec["package"])
     os.makedirs(os.path.join(self.workdir, links_path), exist_ok=True)
 
     # Remove existing symlinks: we'll fetch the ones from the remote next.
@@ -480,8 +599,7 @@ class Boto3RemoteSync:
           target = b"../../" + target
         target = os.fsdecode(target)
         link_path = os.path.join(self.workdir, links_path, os.fsdecode(link_name))
-        dieOnError(execute("ln -sf {} {}".format(target, link_path)),
-                  "Unable to create symlink {} -> {}".format(link_name, target))
+        symlink(target, link_path)
         n_symlinks += 1
       debug("Got %d entries in manifest", n_symlinks)
 
@@ -496,46 +614,87 @@ class Boto3RemoteSync:
       target = os.fsdecode(resp["Body"].read()).rstrip("\n")
       if not target.startswith("../../"):
         target = "../../" + target
-      dieOnError(execute("ln -sf {} {}".format(target, link_path)),
-                 "Unable to create symlink {} -> {}".format(link_key, target))
+      symlink(target, link_path)
 
-  def syncToRemote(self, p, spec):
+  def upload_symlinks_and_tarball(self, spec) -> None:
     if not self.writeStore:
       return
-    tarballNameWithRev = ("{package}-{version}-{revision}.{architecture}.tar.gz"
-                          .format(architecture=self.architecture, **spec))
+
+    dist_symlinks = {}
+    for link_dir in ("dist", "dist-direct", "dist-runtime"):
+      link_dir = "TARS/{arch}/{link_dir}/{package}/{package}-{version}-{revision}" \
+        .format(arch=self.architecture, link_dir=link_dir, **spec)
+
+      debug("Comparing dist symlinks against S3 from %s", link_dir)
+
+      symlinks = []
+      for fname in os.listdir(os.path.join(self.workdir, link_dir)):
+        link_key = os.path.join(link_dir, fname)
+        path = os.path.join(self.workdir, link_key)
+        if os.path.islink(path):
+          hash_path = re.sub(r"^(\.\./)*", "", os.readlink(path))
+          symlinks.append((link_key, hash_path))
+
+      # To make sure there are no conflicts, see if anything already exists in
+      # our symlink directory.
+      symlinks_existing = frozenset(self._s3_listdir(link_dir))
+
+      # If all the symlinks we would upload already exist, skip uploading. We
+      # probably just downloaded a prebuilt package earlier, and it already has
+      # symlinks available.
+      if all(link_key in symlinks_existing for link_key, _ in symlinks):
+        debug("All %s symlinks already exist on S3, skipping upload", link_dir)
+        continue
+
+      # Excluding our own symlinks (above), if there is anything in our link_dir
+      # on the remote, something else is uploading symlinks (or already has)!
+      dieOnError(symlinks_existing,
+                 "Conflicts detected in %s on S3; aborting: %s" %
+                 (link_dir, ", ".join(sorted(symlinks_existing))))
+
+      dist_symlinks[link_dir] = symlinks
+
+    tarball = "{package}-{version}-{revision}.{architecture}.tar.gz" \
+      .format(architecture=self.architecture, **spec)
     tar_path = os.path.join(resolve_store_path(self.architecture, spec["hash"]),
-                            tarballNameWithRev)
-    link_path = os.path.join(resolve_links_path(self.architecture, p),
-                             tarballNameWithRev)
+                            tarball)
+    link_path = os.path.join(resolve_links_path(self.architecture, spec["package"]),
+                             tarball)
     tar_exists = self._s3_key_exists(tar_path)
     link_exists = self._s3_key_exists(link_path)
     if tar_exists and link_exists:
-      debug("%s exists on S3 already, not uploading", tarballNameWithRev)
+      debug("%s exists on S3 already, not uploading", tarball)
       return
-    if tar_exists or link_exists:
-      warning("%s exists already but %s does not, overwriting!",
-              tar_path if tar_exists else link_path,
-              link_path if tar_exists else tar_path)
-    debug("Uploading tarball and symlink for %s %s-%s (%s) to S3",
-          p, spec["version"], spec["revision"], spec["hash"])
-    self.s3.upload_file(Bucket=self.writeStore, Key=tar_path,
-                        Filename=os.path.join(self.workdir, tar_path))
+    dieOnError(tar_exists or link_exists,
+               "%s already exists on S3 but %s does not, aborting!" %
+               (tar_path if tar_exists else link_path,
+                link_path if tar_exists else tar_path))
+
+    debug("Uploading tarball and symlinks for %s %s-%s (%s) to S3",
+          spec["package"], spec["version"], spec["revision"], spec["hash"])
+
+    # Upload the smaller file first, so that any parallel uploads are more
+    # likely to find it and fail.
     self.s3.put_object(Bucket=self.writeStore, Key=link_path,
                        Body=os.readlink(os.path.join(self.workdir, link_path))
                               .lstrip("./").encode("utf-8"))
 
-  def syncDistLinksToRemote(self, link_dir):
-    if not self.writeStore:
-      return
-    for fname in os.listdir(os.path.join(self.workdir, link_dir)):
-      link_key = os.path.join(link_dir, fname)
-      path = os.path.join(self.workdir, link_key)
-      if not os.path.islink(path) or self._s3_key_exists(link_key):
-        continue
-      hash_path = re.sub(r"^(\.\./)*", "", os.readlink(path))
-      self.s3.put_object(Bucket=self.writeStore,
-                         Key=link_key,
-                         Body=os.fsencode(hash_path),
-                         ACL="public-read",
-                         WebsiteRedirectLocation=hash_path)
+    # Second, upload dist symlinks. These should be in place before the main
+    # tarball, to avoid races in the publisher.
+    start_time = time.time()
+    for link_dir, symlinks in dist_symlinks.items():
+      for link_key, hash_path in symlinks:
+        self.s3.put_object(Bucket=self.writeStore,
+                           Key=link_key,
+                           Body=os.fsencode(hash_path),
+                           ACL="public-read",
+                           WebsiteRedirectLocation=hash_path)
+      debug("Uploaded %d dist symlinks to S3 from %s",
+            len(symlinks), link_dir)
+    end_time = time.time()
+    debug("Uploaded %d dist symlinks in %.2f seconds",
+          sum(len(symlinks) for _, symlinks in dist_symlinks.items()),
+          end_time - start_time)
+
+    self.s3.upload_file(Bucket=self.writeStore, Key=tar_path,
+                        Filename=os.path.join(self.workdir, tar_path))
