@@ -1,30 +1,91 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 import yaml
 from os.path import exists
 import hashlib
 from glob import glob
-from os.path import basename, join
+from os.path import basename, join, isdir, islink
 import sys
 import os
 import re
-from datetime import datetime
-try:
-  from collections import OrderedDict
-except ImportError:
-  from ordereddict import OrderedDict
+import platform
 
-from alibuild_helpers.cmd import decode_with_fallback, getoutput
+from datetime import datetime
+from collections import OrderedDict
+from shlex import quote
+from typing import Optional
+
+from alibuild_helpers.cmd import getoutput
 from alibuild_helpers.git import git
-from alibuild_helpers.log import dieOnError
+from alibuild_helpers.log import warning, dieOnError
 
 
 class SpecError(Exception):
   pass
 
+
+def call_ignoring_oserrors(function, *args, **kwargs):
+  try:
+    return function(*args, **kwargs)
+  except OSError:
+    return None
+
+
+def symlink(link_target, link_name):
+  """Match the behaviour of `ln -nsf LINK_TARGET LINK_NAME`, without having to fork.
+
+  Create a new symlink named LINK_NAME pointing to LINK_TARGET. If LINK_NAME
+  is a directory, create a symlink named basename(LINK_TARGET) inside it.
+  """
+  # If link_name is a symlink pointing to a directory, isdir() will return True.
+  if isdir(link_name) and not islink(link_name):
+    link_name = join(link_name, basename(link_target))
+  call_ignoring_oserrors(os.unlink, link_name)
+  os.symlink(link_target, link_name)
+
+
 asList = lambda x : x if type(x) == list else [x]
 
-def star():
-  return re.sub("build.*$", "", basename(sys.argv[0]).lower())
+
+def topological_sort(specs):
+  """Topologically sort specs so that dependencies come before the packages that depend on them.
+
+  This function returns a generator, yielding package names in order.
+
+  The algorithm used here was adapted from:
+  http://www.stoimen.com/blog/2012/10/01/computer-algorithms-topological-sort-of-a-graph/
+  """
+  edges = [(spec["package"], dep) for spec in specs.values() for dep in spec["requires"]]
+  leaves = [spec["package"] for spec in specs.values() if not spec["requires"]]
+  while leaves:
+    current_package = leaves.pop(0)
+    yield current_package
+    # Find every package that depends on the current one.
+    new_leaves = {pkg for pkg, dep in edges if dep == current_package}
+    # Stop blocking packages that depend on the current one...
+    edges = [(pkg, dep) for pkg, dep in edges if dep != current_package]
+    # ...but keep blocking those that still depend on other stuff!
+    leaves.extend(new_leaves - {pkg for pkg, _ in edges})
+  # If we have any edges left, we have a cycle
+  if edges:
+    # Find a cycle by following dependencies
+    cycle = []
+    start = edges[0][0]  # Start with any remaining package
+    current = start
+    max_iter = 10000 # Prevent infinite loops
+    while max_iter > 0:
+      max_iter -= 1
+      cycle.append(current)
+      # Find what current depends on
+      for pkg, dep in edges:
+        if pkg == current:
+          current = dep
+          break
+      if current in cycle:  # We found a cycle
+        cycle = cycle[cycle.index(current):]  # Trim to just the cycle
+        dieOnError(True, "Dependency cycle detected: " + " -> ".join(cycle + [cycle[0]]))
+      if current == start:  # We've gone full circle
+        raise RuntimeError("Internal error: cycle detection failed")
+    assert False, "Unreachable error: cycle detection failed"
 
 
 def resolve_store_path(architecture, spec_hash):
@@ -45,6 +106,18 @@ def resolve_links_path(architecture, package):
   return "/".join(("TARS", architecture, package))
 
 
+def short_commit_hash(spec):
+  """Shorten the spec's commit hash to make it more human-readable.
+
+  This is complicated by the fact that the commit_hash property is not
+  necessarily a commit hash, but might be a tag name. If it is a tag name,
+  return it as-is, else assume it is actually a commit hash and shorten it.
+  """
+  if spec["tag"] == spec["commit_hash"]:
+    return spec["commit_hash"]
+  return spec["commit_hash"][:10]
+
+
 # Date fields to substitute: they are zero-padded
 now = datetime.now()
 nowKwds = { "year": str(now.year),
@@ -67,34 +140,38 @@ def resolve_version(spec, defaults, branch_basename, branch_stream):
   - %(day)s
   - %(hour)s
 
-  with the calculated content"""
+  with the calculated content.
+  """
   defaults_upper = defaults != "release" and "_" + defaults.upper().replace("-", "_") or ""
   commit_hash = spec.get("commit_hash", "hash_unknown")
   tag = str(spec.get("tag", "tag_unknown"))
-  return format(spec["version"],
-                commit_hash=commit_hash,
-                short_hash=commit_hash[0:10],
-                tag=tag,
-                branch_basename = branch_basename,
-                branch_stream = branch_stream or tag,
-                tag_basename=basename(tag),
-                defaults_upper=defaults_upper,
-                **nowKwds)
+  return spec["version"] % {
+    "commit_hash": commit_hash,
+    "short_hash": commit_hash[0:10],
+    "tag": tag,
+    "branch_basename": branch_basename,
+    "branch_stream": branch_stream or tag,
+    "tag_basename": basename(tag),
+    "defaults_upper": defaults_upper,
+    **nowKwds,
+  }
 
 def resolve_tag(spec):
-  """Expand the version replacing the following keywords:
+  """Expand the tag, replacing the following keywords:
   - %(year)s
   - %(month)s
   - %(day)s
-  - %(hour)s"""
-  return format(spec["tag"], **nowKwds)
+  - %(hour)s
+  """
+  return spec["tag"] % nowKwds
+
 
 def normalise_multiple_options(option, sep=","):
   return [x for x in ",".join(option).split(sep) if x]
 
 def prunePaths(workDir):
   for x in ["PATH", "LD_LIBRARY_PATH", "DYLD_LIBRARY_PATH"]:
-    if not x in os.environ:
+    if x not in os.environ:
       continue
     workDirEscaped = re.escape("%s" % workDir) + "[^:]*:?"
     os.environ[x] = re.sub(workDirEscaped, "", os.environ[x])
@@ -107,12 +184,12 @@ def validateSpec(spec):
     raise SpecError("Empty recipe.")
   if type(spec) != OrderedDict:
     raise SpecError("Not a YAML key / value.")
-  if not "package" in spec:
+  if "package" not in spec:
     raise SpecError("Missing package field in header.")
 
 # Use this to check if a given spec is compatible with the given default
 def validateDefaults(finalPkgSpec, defaults):
-  if not "valid_defaults" in finalPkgSpec:
+  if "valid_defaults" not in finalPkgSpec:
     return (True, "", [])
   validDefaults = asList(finalPkgSpec["valid_defaults"])
   nonStringDefaults = [x for x in validDefaults if not type(x) == str]
@@ -126,22 +203,20 @@ def validateDefaults(finalPkgSpec, defaults):
                    "\n".join([" - " + x for x in validDefaults])), validDefaults)
 
 
-def format(s, **kwds):
-  return decode_with_fallback(s) % kwds
-
-
 def doDetectArch(hasOsRelease, osReleaseLines, platformTuple, platformSystem, platformProcessor):
   if platformSystem == "Darwin":
-    import platform
-    if platform.machine() == "x86_64":
-      return "osx_x86-64"
-    else:
-      return "osx_arm64"
+    processor = platformProcessor
+    if not processor:
+      if platform.machine() == "x86_64":
+        processor = "x86-64"
+      else:
+        processor = "arm64"
+    return "osx_%s" % processor.replace("_", "-")
   distribution, version, flavour = platformTuple
   distribution = distribution.lower()
   # If platform.dist does not return something sensible,
   # let's try with /etc/os-release
-  if distribution not in ["ubuntu", "redhat", "centos"] and hasOsRelease:
+  if distribution not in ["ubuntu", "red hat enterprise linux", "redhat", "centos", "almalinux", "rocky linux"] and hasOsRelease:
     for x in osReleaseLines:
       key, is_prop, val = x.partition("=")
       if not is_prop:
@@ -161,7 +236,7 @@ def doDetectArch(hasOsRelease, osReleaseLines, platformTuple, platformSystem, pl
     if version in debian_ubuntu:
       distribution = "ubuntu"
       version = debian_ubuntu[version]
-  elif distribution in ["redhat", "centos"]:
+  elif distribution in ["redhat", "red hat enterprise linux", "centos", "almalinux", "rocky linux"]:
     distribution = "slc"
 
   processor = platformProcessor
@@ -190,7 +265,6 @@ def detectArch():
     osReleaseLines = []
     hasOsRelease = False
   try:
-    import platform
     if platform.system() == "Darwin":
       if platform.machine() == "x86_64":
         return "osx_x86-64"
@@ -199,7 +273,7 @@ def detectArch():
   except:
     pass
   try:
-    import platform, distro
+    import distro
     platformTuple = distro.linux_distribution()
     platformSystem = platform.system()
     platformProcessor = platform.processor()
@@ -209,21 +283,29 @@ def detectArch():
   except:
     return doDetectArch(hasOsRelease, osReleaseLines, ["unknown", "", ""], "", "")
 
-def filterByArchitecture(arch, requires):
+def filterByArchitectureDefaults(arch, defaults, requires):
   for r in requires:
     require, matcher = ":" in r and r.split(":", 1) or (r, ".*")
+    if matcher.startswith("defaults="):
+      wanted = matcher[len("defaults="):]
+      if re.match(wanted, defaults):
+        yield require
     if re.match(matcher, arch):
       yield require
 
-def disabledByArchitecture(arch, requires):
+def disabledByArchitectureDefaults(arch, defaults, requires):
   for r in requires:
     require, matcher = ":" in r and r.split(":", 1) or (r, ".*")
-    if not re.match(matcher, arch):
+    if matcher.startswith("defaults="):
+      wanted = matcher[len("defaults="):]
+      if not re.match(wanted, defaults):
+        yield require
+    elif not re.match(matcher, arch):
       yield require
 
 def readDefaults(configDir, defaults, error, architecture):
-  defaultsFilename = "%s/defaults-%s.sh" % (configDir, defaults)
-  if not exists(defaultsFilename):
+  defaultsFilename = resolveDefaultsFilename(defaults, configDir)
+  if not defaultsFilename:
     error("Default `%s' does not exists. Viable options:\n%s" %
           (defaults or "<no defaults specified>",
            "\n".join("- " + basename(x).replace("defaults-", "").replace(".sh", "")
@@ -245,8 +327,8 @@ def readDefaults(configDir, defaults, error, architecture):
     defaultsBody += "\n# Architecture defaults\n" + archBody
   return (defaultsMeta, defaultsBody)
 
-# Get the appropriate recipe reader depending on th filename
-def getRecipeReader(url, dist=None):
+
+def getRecipeReader(url:str , dist=None):
   m = re.search(r'^dist:(.*)@([^@]+)$', url)
   if m and dist:
     return GitReader(url, dist)
@@ -255,14 +337,14 @@ def getRecipeReader(url, dist=None):
 
 # Read a recipe from a file
 class FileReader(object):
-  def __init__(self, url):
+  def __init__(self, url) -> None:
     self.url = url
   def __call__(self):
     return open(self.url).read()
 
 # Read a recipe from a git repository using git show.
 class GitReader(object):
-  def __init__(self, url, configDir):
+  def __init__(self, url, configDir) -> None:
     self.url, self.configDir = url, configDir
   def __call__(self):
     m = re.search(r'^dist:(.*)@([^@]+)$', self.url)
@@ -270,10 +352,10 @@ class GitReader(object):
     err, d = git(("show", "{gh}:{fn}.sh".format(gh=gh, fn=fn.lower())),
                  directory=self.configDir)
     if err:
-      raise RuntimeError(format("Cannot read recipe %(fn)s from reference %(gh)s.\n" +
-                                "Make sure you run first (this will not alter your recipes):\n" +
-                                "  cd %(dist)s && git remote update -p && git fetch --tags",
-                                dist=self.configDir, gh=gh, fn=fn))
+      raise RuntimeError("Cannot read recipe {fn} from reference {gh}.\n"
+                         "Make sure you run first (this will not alter your recipes):\n"
+                         "  cd {dist} && git remote update -p && git fetch --tags"
+                         .format(dist=self.configDir, gh=gh, fn=fn))
     return d
 
 def yamlLoad(s):
@@ -317,7 +399,7 @@ def parseRecipe(reader):
     err = "Unable to parse %s\n%s" % (reader.url, str(e))
   except yaml.parser.ParserError as e:
     err = "Unable to parse %s\n%s" % (reader.url, str(e))
-  except ValueError as e:
+  except ValueError:
     err = "Unable to parse %s. Header missing." % reader.url
   return err, spec, recipe
 
@@ -346,33 +428,89 @@ def parseDefaults(disable, defaultsGetter, log):
     overrides[f] = dict(**(v or {}))
   return (None, overrides, taps)
 
+def checkForFilename(taps: dict, pkg: str, d: str):
+  return taps.get(pkg, join(d, f"{pkg}.sh"))
+
+def getPkgDirs(configDir):
+  configPath = os.environ.get("BITS_PATH", "").rstrip(":") + ":"
+  pkgDirs = [join(configDir, d) for d in configPath.lstrip(":").split(":")]
+  return pkgDirs
+
+def resolveFilename(taps: dict, pkg: str, configDir: str):
+  for d in getPkgDirs(configDir):
+    filename = checkForFilename(taps, pkg, d)
+    if os.path.exists(filename):
+      return (filename, os.path.abspath(d))
+  return (None, None)
+
+def resolveDefaultsFilename(defaults, configDir) -> Optional[str]:
+  for d in getPkgDirs(configDir):
+    filename = join(d, f"defaults-{defaults}.sh")
+    if os.path.exists(filename):
+      return filename
+
 def getPackageList(packages, specs, configDir, preferSystem, noSystem,
                    architecture, disable, defaults, performPreferCheck, performRequirementCheck,
-                   performValidateDefaults, overrides, taps, log, force_rebuild=()):
+                   performValidateDefaults, overrides, taps: dict, log, force_rebuild=()):
   systemPackages = set()
   ownPackages = set()
   failedRequirements = set()
   testCache = {}
   requirementsCache = {}
+  trackingEnvCache = {}
   packages = packages[:]
   validDefaults = []  # empty list: all OK; None: no valid default; non-empty list: list of valid ones
   while packages:
     p = packages.pop(0)
-    if p in specs:
+    if p in specs or (p == "defaults-release" and ("defaults-" + defaults) in specs):
       continue
-    lowerPkg = p.lower()
-    filename = taps.get(lowerPkg, "%s/%s.sh" % (configDir, lowerPkg))
+
+    # We rewrite all defaults to "defaults-release", so load the correct
+    # defaults package here.
+    # The reason for this rewriting is (I assume) so that packages that are
+    # not overridden by some defaults can be shared with other defaults, since
+    # they will end up with the same hash. The defaults must be called
+    # "defaults-release" for this to work, since the defaults are a dependency
+    # and all dependencies' names go into a package's hash.
+    pkg_filename = ("defaults-" + defaults) if p == "defaults-release" else p.lower()
+
+    filename, pkgdir = resolveFilename(taps, pkg_filename, configDir)
+
+    dieOnError(not filename, "Package %s not found in %s" % (p, configDir))
+    assert(filename is not None)
+
     err, spec, recipe = parseRecipe(getRecipeReader(filename, configDir))
     dieOnError(err, err)
-    dieOnError(spec["package"].lower() != lowerPkg,
+    # Unless there was an error, both spec and recipe should be valid.
+    # otherwise the error should have been caught above.
+    assert(spec is not None)
+    assert(recipe is not None)
+    dieOnError(spec["package"].lower() != pkg_filename,
                "%s.sh has different package field: %s" % (p, spec["package"]))
+    spec["pkgdir"] = pkgdir
+
+    if p == "defaults-release":
+      # Re-rewrite the defaults' name to "defaults-release". Everything auto-
+      # depends on "defaults-release", so we need something with that name.
+      spec["package"] = "defaults-release"
+
+      # Never run the defaults' recipe, to match previous behaviour.
+      # Warn if a non-trivial recipe is found (i.e., one with any non-comment lines).
+      for line in map(str.strip, recipe.splitlines()):
+        if line and not line.startswith("#"):
+          warning("%s.sh contains a recipe, which will be ignored", pkg_filename)
+      recipe = ""
+
     dieOnError(spec["package"] != p,
                "%s should be spelt %s." % (p, spec["package"]))
 
     # If an override fully matches a package, we apply it. This means
     # you can have multiple overrides being applied for a given package.
     for override in overrides:
-      if not re.match("^" + override.strip("^$") + "$", lowerPkg):
+      # We downcase the regex in parseDefaults(), so downcase the package name
+      # as well. FIXME: This is probably a bad idea; we should use
+      # re.IGNORECASE instead or just match case-sensitively.
+      if not re.fullmatch(override, p.lower()):
         continue
       log("Overrides for package %s: %s", spec["package"], overrides[override])
       spec.update(overrides.get(override, {}) or {})
@@ -383,27 +521,87 @@ def getPackageList(packages, specs, configDir, preferSystem, noSystem,
     systemRE = spec.get("prefer_system", "(?!.*)")
     try:
       systemREMatches = re.match(systemRE, architecture)
-    except TypeError as e:
+    except TypeError:
       dieOnError(True, "Malformed entry prefer_system: %s in %s" % (systemRE, spec["package"]))
-    if not noSystem and (preferSystem or systemREMatches):
-      cmd = "REQUESTED_VERSION=%s\n" % resolve_version(spec, defaults, "unavailable", "unavailable")
-      cmd += spec.get("prefer_system_check", "false").strip()
-      if not spec["package"] in testCache:
-        testCache[spec["package"]] = performPreferCheck(spec, cmd.strip())
 
+    noSystemList = []
+    if noSystem == "*":
+      noSystemList = [spec["package"]]
+    elif noSystem is not None:
+      noSystemList = noSystem.split(",")
+    systemExcluded = (spec["package"] in noSystemList)
+    allowSystemPackageUpload = spec.get("allow_system_package_upload", False)
+    # Fill the track env with the actual result from executing the script.
+    for env, trackingCode in spec.get("track_env", {}).items():
+      key = spec["package"] + env
+      if key not in trackingEnvCache:
+        status, out = performPreferCheck(spec, trackingCode)
+        dieOnError(status, "Error while executing track_env for {}: {} => {}".format(key, trackingCode, out))
+        trackingEnvCache[key] = out
+      spec["track_env"][env] = trackingEnvCache[key]
+
+    if (not systemExcluded or allowSystemPackageUpload) and  (preferSystem or systemREMatches):
+      requested_version = resolve_version(spec, defaults, "unavailable", "unavailable")
+      cmd = "REQUESTED_VERSION={version}\n{check}".format(
+        version=quote(requested_version),
+        check=spec.get("prefer_system_check", "false"),
+      ).strip()
+      if spec["package"] not in testCache:
+        testCache[spec["package"]] = performPreferCheck(spec, cmd)
       err, output = testCache[spec["package"]]
-
-      if not err:
-        systemPackages.update([spec["package"]])
-        disable.append(spec["package"])
+      if err:
+        # prefer_system_check errored; this means we must build the package ourselves.
+        ownPackages.add(spec["package"])
       else:
-        ownPackages.update([spec["package"]])
+        # prefer_system_check succeeded; this means we should use the system package.
+        match = re.search(r"^alibuild_system_replace:(?P<key>.*)$", output, re.MULTILINE)
+        if not match and systemExcluded:
+          # No replacement spec name given. Fall back to old system package
+          # behaviour and just disable the package.
+          ownPackages.add(spec["package"])
+        elif not match and not systemExcluded:
+          # No replacement spec name given. Fall back to old system package
+          # behaviour and just disable the package.
+          systemPackages.add(spec["package"])
+          disable.append(spec["package"])
+        elif match:
+          # The check printed the name of a replacement; use it.
+          key = match.group("key").strip()
+          replacement = None
+          for replacement_matcher in spec["prefer_system_replacement_specs"]:
+            if re.match(replacement_matcher, key):
+              replacement = spec["prefer_system_replacement_specs"][replacement_matcher]
+              break
+          if replacement:
+            # We must keep the package name the same, since it is used to
+            # specify dependencies.
+            replacement["package"] = spec["package"]
+            # The version is required for all specs. What we put there will
+            # influence the package's hash, so allow the user to override it.
+            replacement.setdefault("version", requested_version)
+            spec = replacement
+            # Allows generalising the version based on the actual key provided
+            spec["version"] = spec["version"].replace("%(key)s", key)
+            # We need the key to inject the version into the replacement recipe later.
+            spec["key"] = key 
+            recipe = replacement.get("recipe", "")
+            # If there's an explicitly-specified recipe, we're still building
+            # the package. If not, aliBuild will still "build" it, but it's
+            # basically instantaneous, so report to the user that we're taking
+            # it from the system.
+            if recipe:
+              ownPackages.add(spec["package"])
+            else:
+              systemPackages.add(spec["package"])
+          else:
+            warning(f"Could not find named replacement spec for {spec['package']}: {key}, "
+                    "falling back to building the package ourselves.")
 
     dieOnError(("system_requirement" in spec) and recipe.strip("\n\t "),
                "System requirements %s cannot have a recipe" % spec["package"])
     if re.match(spec.get("system_requirement", "(?!.*)"), architecture):
       cmd = spec.get("system_requirement_check", "false")
-      if not spec["package"] in requirementsCache:
+      if spec["package"] not in requirementsCache:
         requirementsCache[spec["package"]] = performRequirementCheck(spec, cmd.strip())
 
       err, output = requirementsCache[spec["package"]]
@@ -426,12 +624,12 @@ def getPackageList(packages, specs, configDir, preferSystem, noSystem,
           validDefaults = None  # no valid default works for all current packages
 
     # For the moment we treat build_requires just as requires.
-    fn = lambda what: disabledByArchitecture(architecture, spec.get(what, []))
+    fn = lambda what: disabledByArchitectureDefaults(architecture, defaults, spec.get(what, []))
     spec["disabled"] += [x for x in fn("requires")]
     spec["disabled"] += [x for x in fn("build_requires")]
-    fn = lambda what: filterByArchitecture(architecture, spec.get(what, []))
-    spec["requires"] = [x for x in fn("requires") if not x in disable]
-    spec["build_requires"] = [x for x in fn("build_requires") if not x in disable]
+    fn = lambda what: filterByArchitectureDefaults(architecture, defaults, spec.get(what, []))
+    spec["requires"] = [x for x in fn("requires") if x not in disable]
+    spec["build_requires"] = [x for x in fn("build_requires") if x not in disable]
     if spec["package"] != "defaults-release":
       spec["build_requires"].append("defaults-release")
     spec["runtime_requires"] = spec["requires"]
@@ -442,14 +640,15 @@ def getPackageList(packages, specs, configDir, preferSystem, noSystem,
     spec["tag"] = spec.get("tag", spec["version"])
     spec["version"] = spec["version"].replace("/", "_")
     spec["recipe"] = recipe.strip("\n")
-    spec["force_rebuild"] = spec["package"] in force_rebuild
+    if spec["package"] in force_rebuild:
+      spec["force_rebuild"] = True
     specs[spec["package"]] = spec
     packages += spec["requires"]
   return (systemPackages, ownPackages, failedRequirements, validDefaults)
 
 
 class Hasher:
-  def __init__(self):
+  def __init__(self) -> None:
     self.h = hashlib.sha1()
   def __call__(self, txt):
     if not type(txt) == bytes:
